/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/24/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cassert>
#include <cstdlib>
#include <dirent.h>
#include <unistd.h>
#include <algorithm>

/*   project headers  */
#include "fs-dir.h"
#include "fs-error.h"
#include "fs-posix-api.h"



namespace duct {
  namespace filesystem {
    namespace api = __posix_api;
    
    
    directory_iterator::dir_itr::~dir_itr()
    {
      api::dir_itr_close(this);
    }

    
    directory_iterator::dir_itr * directory_iterator::construct(const path& p, directory_options options, std::error_code *ec)
    {
      /*   fresh start  */
      if(ec) ec->clear();
      
      /*   make an attempt for creating dir_iter   */
      int errval = 0;
      auto retval = api::dir_itr_ctr(p, errval);
      bool skip = (options & directory_options::skip_permissions_denied) != directory_options::none;
      
      if(errval && !skip) {
        if(ec) *ec = std::error_code(errval, std::generic_category());
        else throw filesystem_error("duct::filesystem::directory_iterator", p, std::error_code(errval, std::generic_category()));
      }
      
      return retval;
    }

    
    directory_iterator& directory_iterator::increment(std::error_code *ec)
    {
      /*   fresh start  */
      if(ec) ec->clear();
      
      /*   make an attempt for incrementing   */
      int errval = 0;
      api::dir_itr_incr(_M_imp, errval);
      
      if(errval) {
        if(ec) *ec = std::error_code(errval, std::generic_category());
        else throw filesystem_error("duct::filesystem::directory_iterator::operator++", std::error_code(errval, std::generic_category()));
      }
      
      return *this;
    }
    
    
    /*   Returns: true if push occurs, otherwise false. Always returns false on error.  */
    bool recursive_directory_iterator::recur_dir_itr::push_directory(std::error_code& ec) noexcept
    {
      ec.clear();
      
      //  Discover if the iterator is for a directory that needs to be recursed into,
      //  taking symlinks and options into account.
      
      if((options & directory_options::_detail_no_push) == directory_options::_detail_no_push)
        options &= ~directory_options::_detail_no_push;
      
      else
      {
        file_status symlink_stat;
        
        if((options & directory_options::following_directory_symlink) != directory_options::following_directory_symlink) {
          symlink_stat = stack.top()->symlink_status(ec);
          if(ec) return false;
        }
        
        if((options & directory_options::following_directory_symlink) == directory_options::following_directory_symlink || !is_symlink(symlink_stat))
        {
          file_status stat = stack.top()->status(ec);
          if(ec || !is_directory(stat)) return false;
          
          directory_iterator next(stack.top()->path(), ec);
          if(!ec && next != directory_iterator())
          {
            stack.push(next);
            ++level;
            return true;
          }
        }
      }
      
      return false;
    }
    
    
    void recursive_directory_iterator::recur_dir_itr::increment(std::error_code& ec) noexcept
    {
      std::error_code ec_push_directory;
      
      //  if various conditions are met, push a directory_iterator into the iterator stack
      if(push_directory(ec)) {
        ec.clear();
        return;
      }
      
      //  Do the actual increment operation on the top iterator in the iterator
      //  stack, popping the stack if necessary, until either the stack is empty or a
      //  non-end iterator is reached.
      while(!stack.empty() && ++stack.top() == directory_iterator())
      {
        stack.pop();
        --level;
      }
    }
    
    void recursive_directory_iterator::recur_dir_itr::pop()
    {
      assert(level > 0);
      
      do
      {
        stack.pop();
        --level;
      } while(!stack.empty() && ++stack.top() == directory_iterator());
    }
    
  }
}

