/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/24/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_fs_dir_h__
#define __cxx_fs_dir_h__ 1

/*   std includes   */
#include <stack>
#include <vector>

/*   project headers   */
#include <cxx/filesystem/fs-status.h>


namespace duct {
  namespace filesystem {
   
    /*   Enum class for directory options    */
    enum class directory_options {
      none = 0,
      following_directory_symlink = 1,
      skip_permissions_denied = 2,
      
      /*   for internal use only   */
      _detail_no_push = 4
    };
    
    /*    directory_options are bitmask constants   */
    constexpr directory_options operator|(directory_options x , directory_options y) {
      return static_cast<directory_options>(static_cast<int>(x) | static_cast<int>(y));
    }
    
    constexpr directory_options operator&(directory_options x , directory_options y) {
      return static_cast<directory_options>(static_cast<int>(x) & static_cast<int>(y));
    }
    
    constexpr directory_options operator^(directory_options x , directory_options y) {
      return static_cast<directory_options>(static_cast<int>(x) ^ static_cast<int>(y)); }
    
    constexpr directory_options operator~(directory_options x) {
      return static_cast<directory_options>(~static_cast<int>(x));
    }
    
    constexpr directory_options& operator&=(directory_options& x , directory_options y) {
      x = x & y;
      return x;
    }
    
    constexpr directory_options& operator|=(directory_options& x , directory_options y) {
      x = x | y;
      return x;
    }
    
    constexpr directory_options& operator^=(directory_options & x , directory_options y) {
      x = x ^ y;
      return x;
    }
    
    
    
    /*    Class for representing directory entries   */
    class directory_entry
    {
    public:
      /*   default constructors    */
      directory_entry() = default;
      
      /*    copy   */
      directory_entry(const directory_entry&) = default;
      directory_entry& operator=(const directory_entry&) = default;
     
      /*   move   */
      directory_entry(directory_entry&&) = default;
      directory_entry& operator=(directory_entry&&) = default;
      
      /*   destructor  */
      ~directory_entry() = default;
      
      /*   other constructors   */
      explicit directory_entry(const path& p) : _M_path(p) {}
      
      /*   modifiers   */
      void assign(const path& p) { _M_path = p;}
      
      void replace_filename(const path& p) {
        _M_path.remove_filename();
        _M_path /= p;
      }
      
      /*    observers    */
      const class path& path() const noexcept {
        return _M_path;
      }
      
      operator const class path&() const noexcept {
        return _M_path;
      }
    
      file_status status() const {
        return filesystem::status(_M_path);
      }
    
      file_status status(std::error_code& ec) const noexcept {
        return filesystem::status(_M_path, ec);
      }
      
      file_status symlink_status() const {
        return filesystem::symlink_status(_M_path);
      }
      
      file_status symlink_status(std::error_code& ec) const noexcept {
        return filesystem::symlink_status(_M_path, ec);
      }
      
      /*    comparisons  */
      bool operator==(const directory_entry& rhs) const noexcept { return _M_path == rhs._M_path;}
      bool operator!=(const directory_entry& rhs) const noexcept { return _M_path != rhs._M_path;}
      bool operator< (const directory_entry& rhs) const noexcept { return _M_path < rhs._M_path;}
      bool operator<=(const directory_entry& rhs) const noexcept { return _M_path <= rhs._M_path;}
      bool operator> (const directory_entry& rhs) const noexcept { return _M_path > rhs._M_path;}
      bool operator>=(const directory_entry& rhs) const noexcept { return _M_path >= rhs._M_path;}
      
    private:
      /*    data memebers   */
      class path _M_path;
    };
    
      
      
    /*    director iterator   */
    class directory_iterator
    : public std::iterator<std::input_iterator_tag,
                           directory_entry,
                           std::ptrdiff_t,
                           const directory_entry *,
                           const directory_entry&>
    {
    public:
      /*   strorage class for implementig the iterator   */
      struct dir_itr
      {
        directory_entry dir_entry;
        void *handle = nullptr, *buffer = nullptr;
        
        /*   destructor  */
        ~dir_itr();
      };
      
    public:
      /*   default constructor (end_iter)  */
      directory_iterator() : _M_imp(nullptr) {}
            
      /*  constructors   */
      explicit directory_iterator(const path& p, directory_options options = directory_options::none)
      : _M_imp(construct(p, options, nullptr)) { this->increment(nullptr);}
      
      directory_iterator(const path& p, std::error_code& ec) noexcept
      : _M_imp(construct(p, directory_options::none, &ec)) { this->increment(&ec);}

      directory_iterator(const path& p, directory_options options, std::error_code& ec) noexcept
      : _M_imp(construct(p, options, &ec)) { this->increment(&ec);}
      
      /*   increment operators   */
      directory_iterator& operator++() {
        return increment(nullptr);
      }
      
      directory_iterator operator++(int) {
        directory_iterator tmp(*this);
        this->operator++();
        return tmp;
      }
      
      directory_iterator& increment(std::error_code& ec) {
        return increment(&ec);
      }
     
      /*    dereference   */
      reference operator*() const { return _M_imp->dir_entry;}
      pointer operator->() const { return &(_M_imp->dir_entry);}
      
      /*   comparisons  */
      bool operator==(const directory_iterator& rhs) const {
        return _M_imp == rhs._M_imp;
      }
 
      bool operator!=(const directory_iterator& rhs) const {
        return _M_imp != rhs._M_imp;
      }
      
    private:
      /*    data mamber  */
      std::shared_ptr<dir_itr> _M_imp;
      
      /*   private functions  */
      static dir_itr * construct(const path&, directory_options, std::error_code *);
      directory_iterator& increment(std::error_code *);
    };


    inline directory_iterator begin(const directory_iterator& iter) {
      return iter;
    }
      
    inline directory_iterator end(const directory_iterator&) {
      return directory_iterator();
    }
      

    /*    director iterator   */
    class recursive_directory_iterator
    : public std::iterator<std::input_iterator_tag,
                           directory_entry,
                           std::ptrdiff_t,
                           const directory_entry *,
                           const directory_entry&>
    {
      /*   strorage class for implementig the iterator   */
      struct recur_dir_itr
      {
        /*  constructor   */
        recur_dir_itr(directory_options opt, directory_iterator&& it)
        : options(opt) { stack.push(std::forward<directory_iterator>(it));}
        
        /*  increment the iterator   */
        void increment(std::error_code&) noexcept;
        
        /*  push sub directory into the stack   */
        bool push_directory(std::error_code&) noexcept;
        
        /*  pop the stack   */
        void pop();
        
        /*   data members   */
        int level = 0;
        directory_options options = directory_options::none;
        std::stack<directory_iterator, std::vector<directory_iterator>> stack;
      };
      
    public:
      /*   default constructor (end_iter)  */
      recursive_directory_iterator() : _M_imp(nullptr) {}
      
      /*  constructors   */
      recursive_directory_iterator(const path& p, directory_options opt, std::error_code& ec) noexcept
      {
        directory_iterator it(p, ec);
        
        if(it != directory_iterator()) {
          try { _M_imp = std::make_shared<recur_dir_itr>(opt, std::move(it));}
          catch(std::bad_alloc) { _M_imp = nullptr; ec = {ENOMEM, std::system_category()};}
        }
      }
      
      recursive_directory_iterator(const path& p, std::error_code& ec) noexcept
      {
        directory_iterator it(p, ec);
        
        if(it != directory_iterator()) {
          try { _M_imp = std::make_shared<recur_dir_itr>(directory_options::none, std::move(it));}
          catch(std::bad_alloc) { _M_imp = nullptr; ec = {ENOMEM, std::system_category()};}
        }
      }
      
      recursive_directory_iterator(const path& p, directory_options opt = directory_options::none)
      {
        directory_iterator it(p);
        
        if(it != directory_iterator())
          _M_imp = std::make_shared<recur_dir_itr>(opt, std::move(it));
      }
      
      
      /*   increment operators   */
      recursive_directory_iterator& operator++()
      {
        std::error_code ec;
        _M_imp->increment(ec);
        if(ec) throw filesystem_error("duct::filesystem::recursive_directory_iterator::operator++", ec);
        if(_M_imp->stack.empty()) _M_imp.reset();
        return *this;
      }
      
      recursive_directory_iterator operator++(int) {
        recursive_directory_iterator tmp(*this);
        this->operator++();
        return tmp;
      }
      
      recursive_directory_iterator& increment(std::error_code& ec)
      {
        _M_imp->increment(ec);
        if(_M_imp->stack.empty()) _M_imp.reset();
        return *this;
      }
      
      /*    dereference   */
      reference operator*() const   {
        return *(_M_imp->stack.top());
      }
      
      pointer operator->() const {
        return _M_imp->stack.top().operator->();
      }
      
      /*   comparisons  */
      bool operator==(const recursive_directory_iterator& rhs) const {
        return _M_imp == rhs._M_imp;
      }
      
      bool operator!=(const recursive_directory_iterator& rhs) const {
        return _M_imp != rhs._M_imp;
      }
      
      /*  observers   */
      int depth() const { return _M_imp->level;}
      directory_options options() const { return _M_imp->options;}
      
      bool recursion_pending() const {
        return (_M_imp->options & directory_options::_detail_no_push) != directory_options::_detail_no_push;
      }
      
      /*  modifiers   */
      void diable_recursion_pending() {
        _M_imp->options |= directory_options::_detail_no_push;
      }
      
      void pop() {
        _M_imp->pop();
        if(_M_imp->stack.empty()) _M_imp.reset();
      }
      
    private:
      /*    data mamber  */
      std::shared_ptr<recur_dir_itr> _M_imp;
    };
      
    inline recursive_directory_iterator begin(const recursive_directory_iterator& iter) {
      return iter;
    }
      
    inline recursive_directory_iterator end(const recursive_directory_iterator&) {
      return recursive_directory_iterator();
    }
      
      
    
  }
}

#endif
