/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/26/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std includes   */
#include <cerrno>
#include <cassert>
#include <cstdlib>
#include <algorithm>
#include <fcntl.h>
#include <utime.h>
#include <dirent.h>
#include <stddef.h>

/*   project headers  */
#include "fs-posix-api.h"


namespace duct {
  
  namespace filesystem {
    
    
    int file_info::allocate() noexcept
    {
      /*   allocate memory but don't throw exception
       *   if it fails return the ENOMEM error number
       */
      if((_M_stat = std::malloc(sizeof(struct ::stat))) == nullptr) return errno;
      
      /*  otherwise return 0 indicating there was no error   */
      return 0;
    }
    
    
    void file_info::deallocate() noexcept
    {
      /*  deallocate memeory   */
      if(_M_stat) free(_M_stat);
    }
    
    
    file_info file_info::info(const path& p, int& errval) noexcept
    {
      /*   create the return value  */
      file_info ret;
      
      if((errval = ret.allocate()) != 0) return ret;
      
      /*   use ::stat() to get file info   */
      struct ::stat *sb = static_cast<struct ::stat*>(ret._M_stat);
      
      if(::stat(p.c_str(), sb) != 0) {
        /*   if the file doesn't exist, that is not an error   */
        if(errno == ENOENT || errno == ENOTDIR) {
          ret.file_status::type(file_type::not_found);
          ret.file_status::permissions(perms::none);
          
          return ret;
        }
        
        /*  some real error happened  */
        else {
          errval = errno;
          ret.file_status::type(file_type::none);
          ret.file_status::permissions(perms::unknown);
        }
        
        return ret;
      }
      
      /*    get the permission  */
      ret.file_status::permissions(static_cast<perms>(sb->st_mode) & perms::mask);
      
      /*   check the file type   */
      if(S_ISDIR(sb->st_mode))  ret.file_status::type(file_type::directory);
      else if(S_ISREG(sb->st_mode))  ret.file_status::type(file_type::regular);
      else if(S_ISBLK(sb->st_mode))  ret.file_status::type(file_type::block);
      else if(S_ISCHR(sb->st_mode))  ret.file_status::type(file_type::character);
      else if(S_ISFIFO(sb->st_mode)) ret.file_status::type(file_type::fifo);
      else if(S_ISSOCK(sb->st_mode)) ret.file_status::type(file_type::socket);
      else ret.file_status::type(file_type::unknown);
    
      return ret;
    }
  
    
    file_info file_info::symlink_info(const path& p, int& errval) noexcept
    {
      /*   create the return value  */
      file_info ret;
      
      if((errval = ret.allocate()) != 0) return ret;
      
      /*   use ::stat() to get file info   */
      struct ::stat *sb = static_cast<struct ::stat*>(ret._M_stat);
      
      if(::lstat(p.c_str(), sb) != 0) {
        /*   if the file doesn't exist, that is not an error   */
        if(errno == ENOENT || errno == ENOTDIR) {
          ret.file_status::type(file_type::not_found);
          ret.file_status::permissions(perms::none);
          
          return ret;
        }
        
        /*  some real error happened  */
        else {
          errval = errno;
          ret.file_status::type(file_type::none);
          ret.file_status::permissions(perms::unknown);
        }
        
        return ret;
      }
      
      /*    get the permission  */
      ret.file_status::permissions(static_cast<perms>(sb->st_mode) & perms::mask);
      
      /*   check the file type   */
      if(S_ISDIR(sb->st_mode))  ret.file_status::type(file_type::directory);
      else if(S_ISREG(sb->st_mode))  ret.file_status::type(file_type::regular);
      else if(S_ISLNK(sb->st_mode))  ret.file_status::type(file_type::symlink);
      else if(S_ISBLK(sb->st_mode))  ret.file_status::type(file_type::block);
      else if(S_ISCHR(sb->st_mode))  ret.file_status::type(file_type::character);
      else if(S_ISFIFO(sb->st_mode)) ret.file_status::type(file_type::fifo);
      else if(S_ISSOCK(sb->st_mode)) ret.file_status::type(file_type::socket);
      else ret.file_status::type(file_type::unknown);
                    
      return ret;
    }
    
    std::uintmax_t file_info::hard_link_count() const noexcept
    {
      return static_cast<uintmax_t>(type() == file_type::none || type() == file_type::not_found
                                    ? -1
                                    : static_cast<struct ::stat *>(_M_stat)->st_nlink);
    }
    
    
    std::uintmax_t file_info::file_size() const noexcept
    {
      return static_cast<uintmax_t>(type() == file_type::none || type() == file_type::not_found
                                    ? -1
                                    : static_cast<struct ::stat *>(_M_stat)->st_size);
    }
    
    file_time_type file_info::last_write_time() const noexcept
    {
      return std::chrono::system_clock::from_time_t(type() == file_type::none || type() == file_type::not_found
                                                    ? -1
                                                    : static_cast<struct ::stat *>(_M_stat)->st_mtime);
    }


    bool file_info::equivalent(const file_info& f) const noexcept
    {
      struct ::stat *s1 = static_cast<struct ::stat *>(_M_stat);
      struct ::stat *s2 = static_cast<struct ::stat *>(f._M_stat);

      return (s1->st_dev == s2->st_dev && s1->st_ino == s2->st_ino
              && s1->st_size == s2->st_size && s1->st_mtime == s2->st_mtime);
    }
    
    
    namespace __posix_api {
      
    
      void last_write_time(const path& p, const file_time_type new_time, int& errval) noexcept
      {
        /*  clean start ...   */
        errval = 0;
        
        /*    set the time   */
        struct ::stat sb;
        
        if(::stat(p.c_str(), &sb)) {
          errval = errno;
          return;
        }
        
        ::utimbuf buf;
        buf.actime = sb.st_atime;
        buf.modtime = std::chrono::system_clock::to_time_t(new_time);
        
        if(::utime(p.c_str(), &buf)) errval = errno;
      }

      
      path current_path(int& errval) noexcept
      {
        /*  clean start ...   */
        errval = 0;
        
        /*   try to get the full path name   */
        for(std::size_t path_max = 256;; path_max *= 2)
        {
          std::unique_ptr<char[]> buf(new(std::nothrow) char[path_max]);
          
          if(!buf) {
            errval = ENOMEM;
            return path{};
          }

          if(::getcwd(buf.get(), static_cast<std::size_t>(path_max)) == nullptr) {
            if(errno != ERANGE && errno != 0) {
              errval = errno;
              return path{};
            }
          } else return path{buf.get()};
        }
      }
      
      
      path read_symlink(const path& p, int& errval) noexcept
      {
        /*   clean start   */
        errval = 0;
        
        /*   use ::lstat() to get file info   */
        struct ::stat sb;
        
        if(::lstat(p.c_str(), &sb) || !S_ISLNK(sb.st_mode)) {
          errval = errno;
          return path{};
        }
        
        /*   allocate memory for buffer   */
        std::unique_ptr<char[]> buf(new(std::nothrow) char[sb.st_size + 1]);
        
        if(!buf) {
          errval = ENOMEM;
          return path{};
        }
        
        /*   read the link name   */
        auto count = ::readlink(p.c_str(), buf.get(), sb.st_size + 1);
        
        if(count == -1) {
          errval = errno;
          return path{};
        }
        
        return path(buf.get(), buf.get() + count);
      }

      
      bool create_directory(const path& p, int& errval) noexcept
      {
        /*  clean start ...   */
        errval = 0;
        
        /*   make an attempt for creating directory   */
        if(0 == ::mkdir(p.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH))
          return true;
        
        /*  there was some error   */
        errval = errno;
   
        /*   the file may be already there   */
        int dummy = 0;
        
        if(errval == EEXIST && is_directory(file_info::info(p, dummy)))
          errval = 0;
        
        return false;
      }

      bool create_directory(const path& p, const path& from, int& errval) noexcept
      {
        /*  clean start ...   */
        errval = 0;
 
        /*   read permissions of `from`   */
        struct ::stat from_stat;
        
        if(::stat(from.c_str(), &from_stat)) {
          errval = errno;
          return false;
        }
        
        /*   make an attempt for creating directory   */
        if(0 == ::mkdir(p.c_str(), from_stat.st_mode)) return true;
        
        /*  there was some error   */
        errval = errno;
        
        /*   the file may be already there   */
        int dummy = 0;
        
        if(errval == EEXIST && is_directory(file_info::info(p, dummy)))
          errval = 0;
          
        return false;
      }

      
      bool remove(const path& p, file_type type, int& errval) noexcept
      {
        /*  clean start ...   */
        errval = 0;
        
        /*   there was some error   */
        if(type == file_type::none) return false;
        
        /*   file doesn't exists, we don't have to delete it   */
        if(type == file_type::not_found) {
          errval = 0;
          return false;
        }
        
        /*   if it is a directory   */
        if(type == file_type::directory) {
          if(::rmdir(p.c_str())) {
            errval = errno;
            return false;
          }
          
          return true;
        }
        
        /*   otherwise   */
        if(::unlink(p.c_str())) {
          errval = errno;
          return false;
        }
        
        return true;
      }
      

      void copy_file(const std::string& from, const std::string& to, int& errval) noexcept
      {
        /*   fresh start ...  */
        errval = 0;
        
        /*   make a copy function   */
        constexpr const std::size_t buf_sz = 32768;
        std::unique_ptr<char[]> buf(new(std::nothrow) char[buf_sz]);
        int infile = -1, outfile = -1;
        
        if(!buf) {
          errval = ENOMEM;
          return;
        }
 
        /*   open the source file   */
        if((infile = ::open(from.c_str(), O_RDONLY)) < 0) {
          errval = infile;
          return;
        }
    
        struct ::stat from_stat;
        if(::stat(from.c_str(), &from_stat) != 0) {
          errval = errno;
          ::close(infile);
          return;
        }
        
        int oflag = O_CREAT | O_WRONLY | O_TRUNC;
        
        /*   open destination file   */
        if((outfile = ::open(to.c_str(), oflag, from_stat.st_mode)) < 0) {
          errval = errno;
          ::close(infile);
          return;
        }
        
        /*   let's do the copy  */
        ssize_t sz, sz_read = 1, sz_write;
        
        while(sz_read > 0 && (sz_read = ::read(infile, buf.get(), buf_sz)) > 0)
        {
          sz_write = 0;
          
          do {
            if((sz = ::write(outfile, buf.get() + sz_write, sz_read - sz_write)) < 0) {
              sz_read = sz;
              break;
            }
            sz_write += sz;
          } while (sz_write < sz_read);
        }
        
        if(::close(infile) < 0) sz_read = -1;
        if(::close(outfile) < 0) sz_read = -1;
        
        if(sz_read < 0) errval = EIO;
      }

      
      
      void dir_itr_close(directory_iterator::dir_itr *itr)
      {
        if(itr->buffer) std::free(itr->buffer);
        if(itr->handle) ::closedir(static_cast<::DIR *>(itr->handle));
        itr->handle = itr->buffer = nullptr;
      }
      
      
      directory_iterator::dir_itr * dir_itr_ctr(const path& p, int& errval)
      {
        /*   a clean start ...  */
        errval = 0;
        
        /*   check if the path is valid   */
        if(p.empty()) {
          errval = ENOENT;
          return nullptr;
        }
        
        /*   create the underlying object   */
        directory_iterator::dir_itr *it = new(std::nothrow) directory_iterator::dir_itr;
        
        if(!it) {
          errval = ENOMEM;
          return nullptr;
        }
        
        /*   opening the directory   */
        if(!(it->handle = ::opendir(p.c_str()))) {
          errval = errno;
          delete it;
          return nullptr;
        }
        
        /*  calculate the buffer size   */
        auto pmax = ::pathconf(p.c_str(), _PC_NAME_MAX);
        constexpr decltype(pmax) _255 = 255;
        std::size_t len = std::max(sizeof(::dirent), offsetof(struct ::dirent, d_name) + std::max(pmax, _255)) + 1;
 
        /*  allocate memory for the buffer   */
        if(!(it->buffer = std::malloc(len))) {
          errval = ENOMEM;
          delete it;
          return nullptr;
        }
        
        /*  everthing looks OK   */
        it->dir_entry.assign(p / ".");
        
        return it;
      }
      
      
      void dir_itr_incr(std::shared_ptr<directory_iterator::dir_itr>& iter, int& errval)
      {
        /*   a clean start ...  */
        errval = 0;
        
        /*   try to do the increment operation   */
        directory_iterator::dir_itr *it = iter.get();
        
        path::string_type filename;
        dirent *result, *entry = static_cast<::dirent *>(it->buffer);
        
        for(;;) {
          /*   try to read the directory   */
          errval = ::readdir_r(static_cast<::DIR *>(it->handle), entry, &result);
          
          /*  end of stream or error    */
          if(!result || errval) {
            iter.reset();
            return;
          }
          
          /*   still OK   */
          filename = entry->d_name;
          
          if(!(filename[0] == '.' && (filename.size() == 1 || (filename[1] == '.' && filename.size()== 2)))) {
            it->dir_entry.replace_filename(filename);
            return;
          }
        }
      }
      

    }   //  namespace __posix_api
  }   //  namespace filesystem
}   //  namespace  duct


