/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/31/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_cxx_memory_resource_h__
#define __duct_cxx_memory_resource_h__ 1


/*   std includes   */
#include <memory>
#include <utility>
#include <tuple>
#include <cstdlib>
#include <atomic>
#include <algorithm>
#include <cstddef>
#include <type_traits>

/*   project headers   */




namespace duct {
  namespace pmr {
    
    /*    Abstract base class for memory resource
     */
    class memory_resource
    {
    protected:
      /*   private members   */
      static constexpr std::size_t max_align = alignof(std::max_align_t);
      
    public:
      /*   destructor   */
      ~memory_resource() = default;
      
      /*   allocate memory   */
      void * allocate(std::size_t bytes, std::size_t alignment = max_align) {
        return do_allocate(bytes, alignment);
      }
      
      /*   deallocate memory   */
      void  deallocate(void *p, std::size_t bytes, std::size_t alignment = max_align) {
        do_deallocate(p, bytes, alignment);
      }
      
      /*   compare memory recources   */
      bool is_equal(const memory_resource& other) const noexcept {
        return do_is_equal(other);
      }
      
    protected:
      /*    abstract allocate function   */
      virtual void * do_allocate(std::size_t, std::size_t) = 0;
      
      /*   abstract deallocate function   */
      virtual void do_deallocate(void *, std::size_t, std::size_t) = 0;
      
      /*   virtual compare function   */
      virtual bool do_is_equal(const memory_resource&) const noexcept = 0;
      
    private:
      /*    default memory resouce    */
      static std::atomic<memory_resource *> _S_default_resource;
      
      /*   friends    */
      friend memory_resource * set_default_resource(memory_resource *);
      friend memory_resource * get_default_resource();
    };
    
    /*   comparisons    */
    inline bool operator==(const memory_resource& a, const memory_resource& b) {
      return &a == &b || a.is_equal(b);
    }

    inline bool operator!=(const memory_resource& a, const memory_resource& b) {
      return !(a == b);
    }
    
    /*    set/get default memory recource    */
    memory_resource * set_default_resource(memory_resource *);
    memory_resource * get_default_resource();

  
    
    /*  20.8.7.2 uses-allocator construction [allocator.uses.construction] 
     *  Uses-allocator construction with allocator Alloc refers to the construction of an object obj of type T,
     *  using constructor arguments v1, v2, ..., vN of types V1, V2, ...,   VN, respectively, and an allocator 
     *  alloc of type Alloc, according to the following rules: ...
     */
    template <bool _use_alloc, bool _prefix_alloc, bool _suffix_alloc, class _Tp, class _Alloc, class... _Args>
    struct __allocator_uses_construction_impl;

    
    template <bool _prefix_alloc, bool _suffix_alloc, class _Tp, class _Alloc, class... _Args>
    struct __allocator_uses_construction_impl<false, _prefix_alloc, _suffix_alloc, _Tp, _Alloc, _Args...>
    {
      /*   cunstructing the object by placement new   */
      static void construct(_Tp *p, _Alloc&, _Args&&... args) {
        ::new(reinterpret_cast<void *>(p)) _Tp{std::forward<_Args>(args)...};
      }
     
      static constexpr std::tuple<_Args...>&&
      make_argument_tuple(_Alloc&, std::tuple<_Args...>&& x) {
        return std::forward<std::tuple<_Args...>>(x);
      }
    };
    
    
    template <bool _suffix_alloc, class _Tp, class _Alloc, class... _Args>
    struct __allocator_uses_construction_impl<true, true, _suffix_alloc, _Tp, _Alloc, _Args...>
    {
      /*   cunstructing the object by placement new   */
      static void construct(_Tp *p, _Alloc& a, _Args&&... args) {
        ::new(reinterpret_cast<void *>(p)) _Tp(std::allocator_arg, a, std::forward<_Args>(args)...);
      }
      
      static std::tuple<std::allocator_arg_t, _Alloc, _Args...>
      make_argument_tuple(_Alloc& a, std::tuple<_Args...>&& x) {
        return std::tuple_cat(std::make_tuple(std::allocator_arg, a), std::forward<std::tuple<_Args...>>(x));
      }
    };
    

    template <class _Tp, class _Alloc, class... _Args>
    struct __allocator_uses_construction_impl<true, false, true, _Tp, _Alloc, _Args...>
    {
      /*   cunstructing the object by placement new   */
      static void construct(_Tp *p, _Alloc& a, _Args&&... args) {
        ::new(reinterpret_cast<void *>(p)) _Tp(std::forward<_Args>(args)..., a);
      }
      
      static std::tuple<_Args..., _Alloc>
      make_argument_tuple(_Alloc& a, std::tuple<_Args...>&& x) {
        return std::tuple_cat(std::forward<std::tuple<_Args...>>(x), std::make_tuple(a));
      }
    };
    
    
    template <class _Tp, class _Alloc, class... _Args>
    struct __allocator_uses_construction_impl<true, false, false, _Tp, _Alloc, _Args...>
    {
      static_assert(sizeof(_Alloc) == 0, "Type uses an allocator but allocator-aware constructor is missing");
      
      /*   cunstructing the object by placement new   */
      static void construct(_Tp *p, _Alloc& a, _Args&&... args) {
        ::new(reinterpret_cast<void *>(p)) _Tp(std::forward<_Args>(args)...);
      }
      
      static constexpr std::tuple<_Args...>&&
      make_argument_tuple(_Alloc& a, std::tuple<_Args...>&& x) {
        return std::forward<std::tuple<_Args...>>(x);
      }
    };
    
    
    /*   A specialization of class template pmr::polymorphic_allocator conforms to the
     *   Allocator requirements ([allocator.requirements] 17.6.3.5). Constructed with 
     *   different memory resources, different instances of the same specialization of 
     *   pmr::polymorphic_allocator can exhibit entirely different allocation behavior. 
     *   This runtime polymorphism allows objects that use polymorphic_allocator to behave 
     *   as if they used different allocator types at run time even though they use the same 
     *   static allocator type.
     */
    template <class _Tp>
    class polymorphic_allocator
    {
      /*    private types   */
      template <class _Up, class... _Args>
      using allocator_uses_construction
      = __allocator_uses_construction_impl<std::uses_allocator<_Up, memory_resource *>::value,
      std::is_constructible<_Up, std::allocator_arg_t, memory_resource *, _Args...>::value,
      std::is_constructible<_Up, _Args..., memory_resource *>::value, _Up, memory_resource *, _Args...>;
    
    public:
      /*   public types   */
      using value_type = _Tp ;
      
      /*   constructors   */
      polymorphic_allocator(memory_resource *r = nullptr) noexcept
      : _M_resource{r ? r : get_default_resource()} {}
      
      /*   copy    */
      polymorphic_allocator(const polymorphic_allocator&) = default;
      polymorphic_allocator& operator=(const polymorphic_allocator&) = default;
      
      template <class U>
      polymorphic_allocator(const polymorphic_allocator<U>& a) noexcept
      : _M_resource(a.resource()) {}

      /*    allocate and deallocate raw memory   */
      _Tp * allocate(std::size_t n) { return static_cast<_Tp*>(_M_resource->allocate(n*sizeof(_Tp), alignof(_Tp)));}
      void deallocate(_Tp *p, std::size_t n) { _M_resource->deallocate(p, n*sizeof(_Tp), alignof(_Tp));}
      
      /*   constructring the object    */
      template <typename U, typename... _Args>
      void construct(U *p, _Args&&... args) { _M_construct(p, std::forward<_Args>(args)...);}
      
      /*   scoped_allocator support    */
      template <class T1, class T2, class... Args1, class... Args2>
      void construct(std::pair<T1, T2> *p, std::piecewise_construct_t, std::tuple<Args1...> x, std::tuple<Args2...> y)
      {
        _M_construct(p, std::piecewise_construct,
                     allocator_uses_construction<T1, Args1...>::make_argument_tuple(_M_resource, std::move(x)),
                     allocator_uses_construction<T2, Args2...>::make_argument_tuple(_M_resource, std::move(y)));
      }

      template <class T1, class T2>
      void construct(std::pair<T1, T2> *p) {
        _M_construct(p, std::piecewise_construct, std::tuple<>(), std::tuple<>());
      }
 
      template <class T1, class T2, class U, class V>
      void construct(std::pair<T1, T2> *p, U&& x, V&& y) {
        _M_construct(p, std::piecewise_construct,
                     std::forward_as_tuple(std::forward<U>(x)),
                     std::forward_as_tuple(std::forward<V>(y)));
      }
  
      template <class T1, class T2, class U, class V>
      void construct(std::pair<T1, T2> *p, const std::pair<U, V>& x) {
        _M_construct(p, std::piecewise_construct,
                     std::forward_as_tuple(x.first),
                     std::forward_as_tuple(x.second));
      }
   
      template <class T1, class T2, class U, class V>
      void construct(std::pair<T1, T2> *p, std::pair<U, V>&& x) {
        _M_construct(p, std::piecewise_construct,
                     std::forward_as_tuple(std::forward<U>(x.first)),
                     std::forward_as_tuple(std::forward<V>(x.second)));
      }
    
      
      /*    destroy the object   */
      template <typename U>
      void destroy(U *p) { p->~U();}
      
      /*    Return a default-constructed allocator    */
      polymorphic_allocator select_on_container_copy_construction() const {
        return {};
      }
      
      /*    returns the memory recource pointer   */
      memory_resource * resource() const { return _M_resource;}
      
    private:
      /*   pointer to the resource object   */
      memory_resource *_M_resource;
      
      /*   helper funtion for construct functions    */
      template <typename U, typename... _Args>
      void _M_construct(U *p, _Args&&... args) {
        allocator_uses_construction<U, _Args...>::construct(p, _M_resource, std::forward<_Args>(args)...);
      }
    };
    
    /*    comparision of polymorphic allocators   */
    template <class T1, class T2>
    inline bool operator==(const polymorphic_allocator<T1>& a, const polymorphic_allocator<T2>& b) {
      return *a.resource() == *b.resource();
    }
    
    template <class T1, class T2>
    inline bool operator!=(const polymorphic_allocator<T1>& a, const polymorphic_allocator<T2>& b) {
      return !(a == b);
    }

    
    /*  w.x.4.1 resource_adaptor [resource.adaptor.overview]
     *  An instance of resource_adaptor<Allocator> is an adaptor that wraps a memory_resource 
     *  interface around Allocator. In order that resource_adaptor<X<T>> and resource_adaptor<X<U>> 
     *  are the same type for any allocator template X and types T and U, resource_adaptor<Allocator> 
     *  is rendered as an alias to a class template such that Allocator is rebound to a char value type 
     *  in every specialization of the class template. 
     */
    template <class _Alloc>
    class resource_adaptor_imp : public memory_resource
    {
    public:
      /*   underlying allocator type   */
      using allocator_type =  _Alloc;
      
      /*   constructors   */
      resource_adaptor_imp() = default;
      resource_adaptor_imp(resource_adaptor_imp&&) = default;
      resource_adaptor_imp(const resource_adaptor_imp&) = default;
  
      explicit resource_adaptor_imp(_Alloc&& a) : _M_alloc{std::move(a)} {}
      explicit resource_adaptor_imp(const _Alloc&& a) : _M_alloc{std::move(a)} {}
      
      /*    defaulted assignments    */
      resource_adaptor_imp& operator=(resource_adaptor_imp&&) = default;
      resource_adaptor_imp& operator=(const resource_adaptor_imp&) = default;
 
      /*   returns the underlying allocator   */
      allocator_type get_allocator() const { return _M_alloc;}
      
    protected:
      /*   allocate memory (alignment argument is ignored here we allocate raw memory ???)    */
      void * do_allocate(std::size_t bytes, std::size_t alignment) final {
        std::size_t new_size = _S_aligned_size(bytes, _S_supported(alignment) ? alignment : max_align);
        return std::allocator_traits<_Alloc>::allocate(_M_alloc, new_size);
      }
      
      /*   deallocate memory (alignment argument is ignored here we allocate raw memory ???)  */
      void do_deallocate(void *ptr, std::size_t size, std::size_t alignment) final
      {
        using _Traits = std::allocator_traits<_Alloc>;
        
        std::size_t new_size = _S_aligned_size(size, _S_supported(alignment) ? alignment : max_align);
        _Traits::deallocate(_M_alloc, static_cast<typename _Traits::pointer>(ptr), new_size);
      }
      
      /*   comparision   */
      bool do_is_equal(const memory_resource& other) const noexcept final {
        const resource_adaptor_imp *p = dynamic_cast<const resource_adaptor_imp *>(&other);
        return (p ? p->_M_alloc == _M_alloc : false);
      }
      
    private:
      /*   underlying allocator   */
      allocator_type _M_alloc;
  
      // Calculate Aligned Size
      // Returns a size that is larger than or equal to __size and divided by
      // __alignment, where __alignment is required to be the power of 2.
      static std::size_t _S_aligned_size(std::size_t bytes, std::size_t alignment) {
        return ((bytes - 1) | (alignment - 1)) + 1;
      }
      
      // Determine whether alignment meets one of those preconditions:
      // 1. Equals to Zero
      // 2. Is power of two
      static bool _S_supported (std::size_t x) {
        return ((x != 0) && !(x & (x - 1)));
      }
    };


    /*    default resource adaptor   */
    template<class _Alloc>
    using resource_adaptor = resource_adaptor_imp<typename std::allocator_traits<_Alloc>::template rebind_alloc<char>>;
 
    /*   default resource type, it is based on the standard allocator    */
    using new_delete_resource = resource_adaptor<std::allocator<char>>;

    /*   the global default new/delete memory resource    */
    extern new_delete_resource _G_new_delete_resource_singleton;

    /*    get/set the default memory resource   */
    inline memory_resource * get_default_resource()
    {
      memory_resource *ret = memory_resource::_S_default_resource.load();
      if(nullptr == ret) ret = &_G_new_delete_resource_singleton;
      return ret;
    }
    
    inline memory_resource * set_default_resource(memory_resource *r)
    {
      if(nullptr == r) r = &_G_new_delete_resource_singleton;
      return memory_resource::_S_default_resource.exchange(r);
    }


  }
}

#endif
