/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1



/*   std includes   */
#include <cmath>
#include <utility>
#include <fstream>
#include <string>
#include <stack>


/*   project headers   */
#include <bits/mc-shower-history.h>
#include <bits/mc-hard-params.h>
#include <bits/mc-splitting-prob.h>
#include <bits/qcd-color-overlap.h>
#include <bits/mc-engine-opts.h>
#include <cxx/cloned-ptr.h>



namespace duct {
  
  namespace __details {
    
    /*  This function selects the color operator. This can be different in different
     *  shower implementations. This function select the color operator in dipole shower.
     */
    double selecting_color_operator_su3(random_engine_type&, splitting_variables<shower<lo>> *, const particle_state *, const model_ref&);
    double selecting_color_operator_u3(random_engine_type&, splitting_variables<shower<lo>> *, const particle_state *, const model_ref&);
    double selecting_color_operator_su3_ggg(random_engine_type&, splitting_variables<shower<lo>> *, const particle_state *, const model_ref&);
    
    
    template<class _Proc> inline
    double selecting_color_operator_su3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_su3(rne, &(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
    
    template<class _Proc> inline
    double selecting_color_operator_u3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_u3(rne, &(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
 
    
    template<class _Proc> inline
    double selecting_color_operator_su3_ggg(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_su3_ggg(rne, &(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
    
  }    //   namespace __details

  
    
  
  /*  This is the shower evolution code. It uses a given shower probability class to make 
   *  trial emission. An instance of the shower probability class is stored in the this
   *  object. The type of the probability class is erased.
   */
  template<class _Proc>
  class shower_engine
  {
    /*    mimicing concept    */
    template<typename T> static constexpr bool
    __non_self = (!std::is_same<shower_engine, typename std::decay<T>::type>::value
                  && !std::is_base_of<shower_engine, typename std::decay<T>::type>::value);
    
    
    /*   private types   */
    using _History = shower_history<_Proc>;
    using _Iterator = typename _History::iterator;
    using _Branch = std::pair<_Iterator, unsigned>;
   
    
  public:
    /*    default constructor    */
    shower_engine() = default;
    
    /*    create the shower engine from a given shower probability   */
    template<class _ShowerProb, typename _X = typename std::enable_if_t<__non_self<_ShowerProb>>>
    explicit shower_engine(_ShowerProb&& prob) : _M_trial{new obj<_ShowerProb>{std::forward<_ShowerProb>(prob)}} {}
    
    
    /*   shower evolution (at the moment it is just the LC+ approximation)   */
    void evolve(double tmax, shower_history<_Proc>& hist)
    {
      hist.k2T_cutoff = sqr(_M_opts.kT_cutoff);
      hist.max_color_suppression = _M_opts.max_color_suppression;
      _M_evolve_lcplus(hist, tmax);
    }

    /*   same like evolve but it returns the history   */
    shower_history<_Proc>& operator()(double tmax, shower_history<_Proc>& hist) {
      this->evolve(tmax, hist);
      return hist;
    }
    
    /*   Setting/getting engine options     */
    void options(const shower_engine_opts& opts) { _M_opts = opts;}
    const shower_engine_opts& options() const { return _M_opts;}
    
    /*   set the random number seed   */
    void seed(random_engine_type::result_type s) { _M_random.seed(s);}
    
  private:
    /*   random engine   */
    random_engine_type _M_random;
    
    /*   engine parameters    */
    shower_engine_opts _M_opts;
    
    /*   calculates color overlaps in the evolution    */
    color_overlap_su3 _M_overlap_su3;
    color_overlap_u3 _M_overlap_u3;
    
    /*   some workspace for keep tracking of the unshowered nodes   */
    std::stack<_Branch> _M_not_showered;
    
  private:
    /*    make one shower step   */
    _Iterator _M_make_one_step(shower_history<_Proc>&, unsigned int, _Iterator, double);
    
    /*   LC+ approximation of the shower evolution     */
    void _M_evolve_lcplus(shower_history<_Proc>&, double);
    
    
    /*   abstract class for type erasing   */
    struct base {
      virtual bool make_trial(shower_stage<_Proc>&, random_engine_type&, double, double) = 0;
      virtual __details::thrs_weight threshold(const shower_engine_opts::threshold_opts&, double, double) const = 0;
    };
    
    
    /*   this stores an instanse of the actual splitting probability   */
    template<class _ShowerProb>
    struct obj : base
    {
      explicit obj(_ShowerProb&& s) : prob{std::forward<_ShowerProb>(s)} {}
      
      bool make_trial(shower_stage<_Proc>& s, random_engine_type& r, double tmU, double tmL) {
        return prob.make_trial(s, r, tmU, tmL);
      }
      
      __details::thrs_weight threshold(const shower_engine_opts::threshold_opts& opts, double tmH, double tmL) const {
        return prob.threshold(opts, tmH, tmL);
      }
      
      _ShowerProb prob;
    };
    
    
    /*   we put an instance of obj into an cloned_ptr   */
    cloned_ptr<base> _M_trial;
    
    
    /*   change color group (SU(Nc) ==> U(Nc))    */
    void _M_change_color_group(_Iterator next, unsigned int hcsidx)
    {
      if(next->is_color_u3 == false && next->state.color_suppression_index >= hcsidx + _M_opts.max_color_suppression)
      {
        next->weights.unitary.main *= _M_overlap_su3(next->state);
        next->weights.unitary.main /= _M_overlap_u3(next->state);
        next->is_color_u3 = true;
      }
    }
    
  };
  
  
  
  template<class _Proc>
  typename shower_history<_Proc>::iterator shower_engine<_Proc>::
  _M_make_one_step(shower_history<_Proc>& hist, unsigned int hcsidx, _Iterator current, double tmax)
  {
    /*     make trial shower emissions    */
    bool is_possible = _M_trial -> make_trial(*current, _M_random, tmax, current->current_time);
    
    
    /*    calculating the threshold weight    */
    if(is_hadron_collision<_Proc>) {
      if(_M_opts.threshold.enabled)
        current->weights.apply_threshold_correction(_M_trial -> threshold(_M_opts.threshold, current->next_time, current->current_time));
    }

    /*   if there is no splitting we return with the current state    */
    if(!is_possible) return current;
    

    /*   selecting the color operator    */
    const model_ref &mdl = current->state.get_model();
    int cl = mdl.color(current->vars.fl), cm = mdl.color(current->vars.fm);

    
    /*   emitter is a frail gluon and it is a g -> g + g splitting   */
    if(_M_opts.allow_color_branching
       && current->is_color_u3 == false
       && current->state[current->vars.emitter].health == 'f'
       && cl == 8 && cm == 8)
    {
      /*   select the "composite" color operator   */
      double weight = __details::selecting_color_operator_su3_ggg(_M_random, *current);
      
      /*   make the mother state of the new branch    */
      _Iterator branch = hist.append_branch(current, *current);

      /*   fixing color vertices in the mother states   */
      if(current->vars.sigma == 1) {
        current->vars.sigmaP = 1;
        branch->vars.sigmaP = -1;
      }
      
      else if(current->vars.sigma == -1) {
        current->vars.sigmaP = -1;
        branch->vars.sigmaP = 1;
      }
      
      else if(current->vars.sigmaP == 1) {
        current->vars.sigma = 1;
        branch->vars.sigma = -1;
      }
      
      else if(current->vars.sigmaP == -1) {
        current->vars.sigma = -1;
        branch->vars.sigma = 1;
      }

      
      
      /*    appending an uninitialized node to the history     */
      _Iterator next = hist.append(current);
      
      next->weights = current->weights;
      next->weights.unitary.main *= weight;
      
      next->is_color_u3 = current->is_color_u3;
      next->scale = current->vars.scale;
      next->current_time = current->next_time;
      next->next_time = tmax;
      next->vetos = current->vetos;
      
      /*     assigning the new momenta and flavors    */
      next->qnull = __details::shower_mapping_no_color(next->state, current->state, current->vars)(current->qnull);
      
      
      /*   appending the new state on the other branch   */
      _Iterator nextB = hist.append(branch, *next);
      nextB->weights.unitary.main *= -1.0;

      
      /*   mapping the color states    */
      __details::shower_color_mapping(next->state, current->state, current->vars);
      __details::shower_color_mapping(nextB->state, branch->state, branch->vars);

     
      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next, hcsidx);
      _M_change_color_group(nextB, hcsidx);
      
      /*   keep showering with the branch of next and do the evolution for the branch of "branch" later   */
      if(std::abs(nextB->weights.unitary.main) > 0.0)
        _M_not_showered.push(_Branch{nextB, hcsidx});
      
 
      return next;
    }
    
    /*   emitter is a frail gluon and it is a g -> q + q~ splitting   */
    else if(_M_opts.allow_color_branching
            && current->is_color_u3 == false
            && current->state[current->vars.emitter].health == 'f'
            && std::abs(cl) == 3 && cm == -cl)
    {
      /*   make the mother state of the new branch    */
      _Iterator branch[3];
      
      branch[0] = hist.append_branch(current, *current);
      branch[1] = hist.append_branch(current, *current);
      branch[2] = hist.append_branch(current, *current);
      
      
      /*    fixing color vertices in the mother states   */
      int sigma = (cl == 3 ? 2 : 3);

      current->vars.sigma  = sigma;
      current->vars.sigmaP = sigma;
      
      branch[0]->vars.sigma  =  sigma;
      branch[0]->vars.sigmaP = -sigma;
   
      branch[1]->vars.sigma  = -sigma;
      branch[1]->vars.sigmaP =  sigma;
    
      branch[2]->vars.sigma  = -sigma;
      branch[2]->vars.sigmaP = -sigma;
     

      /*    appending an uninitialized node to the history     */
      _Iterator next[4];
      
      next[0] = hist.append(current);
      
      next[0]->weights = current->weights;
      next[0]->is_color_u3 = current->is_color_u3;
      next[0]->scale = current->vars.scale;
      next[0]->current_time = current->next_time;
      next[0]->next_time = tmax;
      next[0]->vetos = current->vetos;
      
      /*   assigning the new momenta and flavors    */
      next[0]->qnull = __details::shower_mapping_no_color(next[0]->state, current->state, current->vars)(current->qnull);
      
      
      /*   appending the new state on the other branch   */
      next[1] = hist.append(branch[0], *(next[0]));
      next[2] = hist.append(branch[1], *(next[0]));
      next[3] = hist.append(branch[2], *(next[0]));
      
      
      /*   fixing the weights   */
      next[0]->weights.unitary.main *=  0.5*qcd::Nc/qcd::Cf;
      next[1]->weights.unitary.main *= -0.5/qcd::Cf;
      next[2]->weights.unitary.main *= -0.5/qcd::Cf;
      next[3]->weights.unitary.main *=  0.5/(qcd::Nc*qcd::Cf);
      
      
      /*   mapping the color states    */
      __details::shower_color_mapping(next[0]->state, current->state, current->vars);
      __details::shower_color_mapping(next[1]->state, branch[0]->state, branch[0]->vars);
      __details::shower_color_mapping(next[2]->state, branch[1]->state, branch[1]->vars);
      __details::shower_color_mapping(next[3]->state, branch[2]->state, branch[2]->vars);
      
      
      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next[0], hcsidx);
      _M_change_color_group(next[1], hcsidx);
      _M_change_color_group(next[2], hcsidx);
      _M_change_color_group(next[3], hcsidx);
      
      
      /*   keep showering with the branch of next and do the evolution for the branch of "branch" later   */
      for(int k : {1,2,3})
        if(std::abs(next[k]->weights.unitary.main) > 0.0)
          _M_not_showered.push(_Branch{next[k], hcsidx});
      
      
      return next[0];
    }
    
    /*   no brancing in the history, we have the simple Markovian chain   */
    else
    {
      /*    selecting color operator    */
      double weight = 0.0;
    
      if(current->is_color_u3) weight = __details::selecting_color_operator_u3(_M_random, *current);
      else weight = __details::selecting_color_operator_su3(_M_random, *current);
    
      /*    appending an uninitialized node to the history     */
      _Iterator next = hist.append(current);
      
      next->weights = current->weights;
      next->weights.unitary.main *= weight;
      
      next->is_color_u3 = current->is_color_u3;
      next->scale = current->vars.scale;
      next->current_time = current->next_time;
      next->next_time = tmax;
      next->vetos = current->vetos;
      
      /*     assigning the new momenta and flavors    */
      next->qnull = __details::shower_mapping(next->state, current->state, current->vars)(current->qnull);
      
      /*   change color group (SU(Nc) ==> U(Nc))    */
      _M_change_color_group(next, hcsidx);
      
      return next;
    }
  }
 
  
  
  
 
  template<class _Proc>
  void shower_engine<_Proc>::_M_evolve_lcplus(shower_history<_Proc>& hist, double tmF)
  {
    /*    save all the leaf node of the tree   */
    for(auto ip = hist.begin(); ip != hist.end(); ++ip)
      if(tmF > ip->next_time && std::abs(ip->weights.unitary.main) > 0.0)
      {
        if(ip->vetos.min_k2T < hist.k2T_cutoff)
          ip->vetos.min_k2T = hist.k2T_cutoff;
        
        /*    change color group of the starting stage, SU(Nc) ==> U(Nc), if it is needed   */
        unsigned int hcsidx = ip.hard().state.color_suppression_index;
 
        _M_change_color_group(ip, hcsidx);

        /*   if the weight is not zero we enroll this stage for shower evolution    */
        if(std::abs(ip->weights.unitary.main) != 0.0)
          _M_not_showered.push(_Branch{ip, hcsidx});
      }
    
    
    /*   shower every stages   */
    while(_M_not_showered.empty() == false)
    {
      /*   pop the top of the stack   */
      _Branch top = _M_not_showered.top(); _M_not_showered.pop();
      
      
      /*    start the evolution   */
      _Iterator current, next = top.first;
      
      do {
        current = next;
        next = _M_make_one_step(hist, top.second, current, tmF);
      } while(current != next && std::abs(next->weights.unitary.main) > 0.0);
    }
  }
    
  

}   /*   namespace duct   */

#endif
