/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 14/2/12.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   standard c++ includes

//   project includes
#include "bits/mc-engine.h"
#include "bits/qcd-color.h"



namespace duct {
  namespace __details {
    
    double selecting_color_operator_su3(random_engine_type& rne, splitting_variables<shower<lo>> *vars, const particle_state *p, const model_ref& mdl)
    {
      int fl = vars->fl, fm = vars->fm;
      int cl = mdl.color(fl), cm = mdl.color(fm);
      
      //----- default weight -----
      double weight = 1.0;
      
      //  spliting e.g.: g -> g + g
      if(cl == 8 && cm == 8)
      {
        /* First we have to calculate the probabilities of the operators.   */
        double w[4] = {qcd::Na/qcd::Nc2, qcd::Na/qcd::Nc2, 1.0/qcd::Nc2, 1.0/qcd::Nc2}, wcross = -qcd::Na;
        
        if(p[vars->emitter].health == 'f') {
          w[0] = w[1] = w[2] = w[3] = 1.0;
          weight = 2.0*qcd::Na/qcd::Nc2;
          wcross = -1.0;
        }
        
        if(vars->spectator != vars->emitter)
        {
          unsigned int chi_p = 0u, chi_m = 0u, chiP_p = 0u, chiP_m = 0u;
          
          if(p[vars->emitter].color.next == vars->spectator) ++chi_p;
          else if(p[vars->emitter].color.prev == vars->spectator) ++chi_m;
          
          if(p[vars->emitter].colorP.next == vars->spectator) ++chiP_p;
          else if(p[vars->emitter].colorP.prev == vars->spectator) ++chiP_m;
          
          w[0] *= chi_p + chiP_p;
          w[1] *= chi_m + chiP_m;
          w[2] *= chi_p + chiP_m;
          w[3] *= chi_m + chiP_p;
        }
        
        /*    selecting the operator   */
        switch(std::discrete_distribution<int>{w, w+4}(rne)) {
          case 0: vars->sigma =  1; vars->sigmaP =  1; break;
          case 1: vars->sigma = -1; vars->sigmaP = -1; break;
          case 2: vars->sigma =  1; vars->sigmaP = -1; weight *= wcross; break;
          case 3: vars->sigma = -1; vars->sigmaP =  1; weight *= wcross; break;
        }
      }
      
      //   splitting e.g.: q -> q + g
      else if(cl == 3 && cm == 8) vars->sigma = vars->sigmaP = 1;
      
      //   splitting e.g.: q~ -> q~ + g
      else if(cl == -3 && cm == 8) vars->sigma = vars->sigmaP = -1;
      
      //  spliting e.g.: g -> q + q~
      else if(std::abs(cl) == 3 && cm == -cl)
      {
        int sigma = (cl == 3 ? 2 : 3);
 
        /* First we have to calculate the probabilities of the operators.   */
        double w[4] = {qcd::Nc2, 1.0, 1.0, 1.0}, wlead = 1.0;
        
        if(p[vars->emitter].health == 'f') {
          w[0] = 1.0;
          weight = 4.0/qcd::Na;
          wlead = qcd::Nc2;
        } else weight = (qcd::Nc2+3.0)/qcd::Na;
    
        
        /*    selecting the operator   */
        switch(std::discrete_distribution<int>{w, w+4}(rne)) {
          case 0: vars->sigma =  sigma; vars->sigmaP =  sigma; weight *=  wlead;   break;
          case 1: vars->sigma =  sigma; vars->sigmaP = -sigma; weight *= -qcd::Nc; break;
          case 2: vars->sigma = -sigma; vars->sigmaP =  sigma; weight *= -qcd::Nc; break;
          case 3: vars->sigma = -sigma; vars->sigmaP = -sigma; break;
        }
      }
      
      //  splitting e.g.: q -> g + q in the initial state
      else if(cl == 8 && cm == 3) vars->sigma = vars->sigmaP = 4;
      
      //  splitting e.g.: q~ -> g + q~ in the initial state
      else if(cl == 8 && cm == -3) vars->sigma = vars->sigmaP = -4;
      
      //   unkown splitting
      else throw "selecting_color_operator_su3 : unable to assign the new color state";
      
      return weight;
    }
    
    
    double selecting_color_operator_u3(random_engine_type& rne, splitting_variables<shower<lo>> *vars, const particle_state *p, const model_ref& mdl)
    {
      int fl = vars->fl, fm = vars->fm;
      int cl = mdl.color(fl), cm = mdl.color(fm);
      
      //----- default weight -----
      double weight = 1.0;
      
      //  spliting e.g.: g -> g + g
      if(cl == 8 && cm == 8)
      {
        /* First we have to calculate the probabilities of the operators.   */
        unsigned int chiP = 0U, chiM = 0U;
        
        if(vars->spectator == vars->emitter)
          chiP = chiM = 1u;
        else
        {
          if(p[vars->emitter].color.next == vars->spectator) ++chiP;
          else if(p[vars->emitter].color.prev == vars->spectator) ++chiM;
          
          if(p[vars->emitter].colorP.next == vars->spectator) ++chiP;
          else if(p[vars->emitter].colorP.prev == vars->spectator) ++chiM;
        }
        
        /*   selecting the color operator    */
        std::uniform_real_distribution<double> rng(0.0,1.0);
        
        if((chiP+chiM)*rng(rne) < static_cast<double>(chiP)) vars->sigma = vars->sigmaP = 1;
        else vars->sigma = vars->sigmaP = -1;
      }
      
      //   splitting e.g.: q -> q + g
      else if(cl == 3 && cm == 8) vars->sigma = vars->sigmaP = 1;
      
      //   splitting e.g.: q~ -> q~ + g
      else if(cl == -3 && cm == 8) vars->sigma = vars->sigmaP = -1;
      
      //  spliting e.g.: g -> q + q~
      else if(cl ==  3 && cm == -cl) vars->sigma = vars->sigmaP = 2;
      else if(cl == -3 && cm == -cl) vars->sigma = vars->sigmaP = 3;
      
      //  splitting e.g.: q -> g + q in the initial state
      else if(cl == 8 && cm == 3) vars->sigma = vars->sigmaP = 4;
      
      //  splitting e.g.: q~ -> g + q~ in the initial state
      else if(cl == 8 && cm == -3) vars->sigma = vars->sigmaP = -4;
      
      //   unkown splitting
      else throw "selecting_color_operator_u3 : unable to assign the new color state";
      
      return weight;
    }
  

    
    
    double selecting_color_operator_su3_ggg(random_engine_type& rne, splitting_variables<shower<lo>> *vars, const particle_state *p, const model_ref& mdl)
    {
      /* First we have to calculate the probabilities of the operators.   */
      double w[4] = {0.0, 0.0, 0.0, 0.0};

      if(vars->spectator == vars->emitter)
        w[0] = w[1] = w[2] = w[3] = 1.0;
      else
      {
        if(p[vars->emitter].color.next == vars->spectator) w[0] = 1.0;
        else if(p[vars->emitter].color.prev == vars->spectator) w[1] = 1.0;
        
        if(p[vars->emitter].colorP.next == vars->spectator) w[2] = 1.0;
        else if(p[vars->emitter].colorP.prev == vars->spectator) w[3] = 1.0;
      }
      
      
      /*    selecting the operator    */
      switch(std::discrete_distribution<int>{w, w+4}(rne))
      {
        case 0: vars->sigma =  1; vars->sigmaP =  0; break;
        case 1: vars->sigma = -1; vars->sigmaP =  0; break;
        case 2: vars->sigma =  0; vars->sigmaP =  1; break;
        case 3: vars->sigma =  0; vars->sigmaP = -1; break;
      }
      
      
      return qcd::Na/qcd::Nc2;
    }
    
  }  //   namespace __details
  

  
}    //   namespace nlo
