/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán qcd::Nagy on 12/27/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>
#include <algorithm>

//   project headers
#include "proc-hhcjets.h"


namespace duct {
  
  
  /*   Table of all the possible subprocesses
   */
  const hhcjets::subprocess hhcjets::_M_proc_tbl[] = {
    //----- process 0 -----
    {{0,0,0,0}, &hhcjets::_M_process_gggg, &hhcjets::_S_select_color_gggg},
    
    /*    Process 1-6: g + g --> q + qbar, where q = d,u,s,c,b,t
     */
    {{0,0,1,-1}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,2,-2}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,3,-3}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,4,-4}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,5,-5}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    {{0,0,6,-6}, &hhcjets::_M_process_ggqQ, &hhcjets::_S_select_color_ggqQ},
    
    /*    Process  7-11: g + q --> g + q, where q = d,u,s,c,b
     */
    {{0,-1,0,1}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-2,0,2}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-3,0,3}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-4,0,4}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    {{0,-5,0,5}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gqgq},
    
    /*    Process 12-16: g + qbar --> g + qbar where q = d,u,s,c,b
     */
    {{0,1,0,-1}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,2,0,-2}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,3,0,-3}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,4,0,-4}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    {{0,5,0,-5}, &hhcjets::_M_process_gqgq, &hhcjets::_S_select_color_gQgQ},
    
    /*    Process 17-21: q + g --> g + q, where q = d,u,s,c,b
     */
    {{-1,0,0,1}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-2,0,0,2}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-3,0,0,3}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-4,0,0,4}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    {{-5,0,0,5}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_qggq},
    
    /*    Process 22-26: qbar + g --> g + qbar, where q = d,u,s,c,b
     */
    {{1,0,0,-1}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{2,0,0,-2}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{3,0,0,-3}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{4,0,0,-4}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    {{5,0,0,-5}, &hhcjets::_M_process_qggq, &hhcjets::_S_select_color_QggQ},
    
    /*    Process 27-31: q + qbar --> g + g, where q = d,u,s,c,b
     */
    {{-1,1,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-2,2,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-3,3,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-4,4,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    {{-5,5,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_qQgg},
    
    /*    Process 32-36: qbar + q --> g + g, where q = d,u,s,c,b
     */
    {{1,-1,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{2,-2,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{3,-3,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{4,-4,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    {{5,-5,0,0}, &hhcjets::_M_process_qQgg, &hhcjets::_S_select_color_Qqgg},
    
    /*    process 37-41: d + dbar --> r + rbar, where r =   u,s,c,b,t
     *    process 42-46: u + ubar --> r + rbar, where r = d,  s,c,b,t
     *    process 47-51: s + sbar --> r + rbar, where r = d,u,  c,b,t
     *    process 52-56: c + cbar --> r + rbar, where r = d,u,s,  b,t
     *    process 57-61: b + bbar --> r + rbar, where r = d,u,s,c,  t
     */
    {{-1,1,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-1,1,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-2,2,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-2,2,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-3,3,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-3,3,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-4,4,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,5,-5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-4,4,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    {{-5,5,1,-1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,2,-2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,3,-3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,4,-4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    {{-5,5,6,-6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_qQrR},
    
    /*    process 62-66: dbar + d --> rbar + r, where r =   u,s,c,b,t
     *    process 67-71: ubar + u --> rbar + r, where r = d,  s,c,b,t
     *    process 72-76: sbar + s --> rbar + r, where r = d,u,  c,b,t
     *    process 77-81: cbar + c --> rbar + r, where r = d,u,s,  b,t
     *    process 81-86: bbar + b --> rbar + r, where r = d,u,s,c,  t
     */
    {{1,-1,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{1,-1,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{2,-2,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{2,-2,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{3,-3,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{3,-3,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{4,-4,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-5,5}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{4,-4,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    {{5,-5,-1,1}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-2,2}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-3,3}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-4,4}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    {{5,-5,-6,6}, &hhcjets::_M_process_qQrR, &hhcjets::_S_select_color_QqRr},
    
    
    /*    Process 87-90:   d + r --> d + r, where r =   u,s,c,b
     *    Process 91-94:   u + r --> u + r, where r = d,  s,c,b
     *    Process 95-98:   s + r --> s + r, where r = d,u,  c,b
     *    Process 99-102:  c + r --> c + r, where r = d,u,s,  b
     *    Process 103-106: b + r --> b + r, where r = d,u,s,c
     */
    {{-1,-2,1,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-3,1,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-4,1,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-1,-5,1,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-2,-1,2,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-3,2,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-4,2,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-2,-5,2,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-3,-1,3,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-2,3,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-4,3,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-3,-5,3,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-4,-1,4,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-2,4,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-3,4,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-4,-5,4,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    {{-5,-1,5,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-2,5,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-3,5,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    {{-5,-4,5,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qrqr},
    
    /*    Process 107-110: dbar + rbar --> dbar + rbar, where r =   u,s,c,b
     *    Process 111-114: ubar + rbar --> ubar + rbar, where r = d,  s,c,b
     *    Process 115-118: sbar + rbar --> sbar + rbar, where r = d,u,  c,b
     *    Process 119-122: cbar + rbar --> cbar + rbar, where r = d,u,s,  b
     *    Process 123-126: bbar + rbar --> bbar + rbar, where r = d,u,s,c
     */
    {{1,2,-1,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,3,-1,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,4,-1,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{1,5,-1,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{2,1,-2,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,3,-2,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,4,-2,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{2,5,-2,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{3,1,-3,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,2,-3,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,4,-3,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{3,5,-3,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{4,1,-4,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,2,-4,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,3,-4,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{4,5,-4,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    {{5,1,-5,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,2,-5,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,3,-5,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    {{5,4,-5,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QRQR},
    
    /*    Process 127-130: d + rbar --> d + rbar, where r =   u,s,c,b
     *    Process 131-134: u + rbar --> u + rbar, where r = d,  s,c,b
     *    Process 135-138: s + rbar --> s + rbar, where r = d,u,  c,b
     *    Process 139-142: c + rbar --> c + rbar, where r = d,u,s,  b
     *    Process 143-146: b + rbar --> b + rbar, where r = d,u,s,c
     */
    {{-1,2,1,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,3,1,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,4,1,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-1,5,1,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-2,1,2,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,3,2,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,4,2,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-2,5,2,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-3,1,3,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,2,3,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,4,3,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-3,5,3,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-4,1,4,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,2,4,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,3,4,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-4,5,4,-5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    {{-5,1,5,-1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,2,5,-2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,3,5,-3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    {{-5,4,5,-4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_qRqR},
    
    /*    Process 147-150: dbar + r --> dbar + r, where r =   u,s,c,b
     *    Process 151-154: ubar + r --> ubar + r, where r = d,  s,c,b
     *    Process 155-158: sbar + r --> sbar + r, where r = d,u,  c,b
     *    Process 159-162: cbar + r --> cbar + r, where r = d,u,s,  b
     *    Process 163-166: bbar + r --> bbar + r, where r = d,u,s,c
     */
    {{1,-2,-1,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-3,-1,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-4,-1,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{1,-5,-1,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{2,-1,-2,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-3,-2,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-4,-2,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{2,-5,-2,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{3,-1,-3,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-2,-3,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-4,-3,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{3,-5,-3,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{4,-1,-4,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-2,-4,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-3,-4,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{4,-5,-4,5}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    {{5,-1,-5,1}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-2,-5,2}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-3,-5,3}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    {{5,-4,-5,4}, &hhcjets::_M_process_qrqr, &hhcjets::_S_select_color_QrQr},
    
    
    /*    Process 167-169:   d + dbar --> d + dbar
     *    Process 170-172:   u + ubar --> u + ubar
     *    Process 173-175:   s + sbar --> s + sbar
     *    Process 176-168:   c + cbar --> c + cbar
     *    Process 179-181:   b + bbar --> b + bbar
     */
    {{-1,1,1,-1}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-1,1,1,-1}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-1,1,1,-1}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-2,2,2,-2}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-2,2,2,-2}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-2,2,2,-2}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-3,3,3,-3}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-3,3,3,-3}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-3,3,3,-3}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-4,4,4,-4}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-4,4,4,-4}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-4,4,4,-4}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    {{-5,5,5,-5}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_qQqQ},
    {{-5,5,5,-5}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_qQqQ},
    {{-5,5,5,-5}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_qQqQ},
    
    /*    Process 182-184:   dbar + d --> dbar + d
     *    Process 185-187:   ubar + u --> ubar + u
     *    Process 188-190:   sbar + s --> sbar + s
     *    Process 191-193:   cbar + c --> cbar + c
     *    Process 194-196:   bbar + b --> bbar + b
     */
    {{1,-1,-1,1}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{1,-1,-1,1}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{1,-1,-1,1}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{2,-2,-2,2}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{2,-2,-2,2}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{2,-2,-2,2}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{3,-3,-3,3}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{3,-3,-3,3}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{3,-3,-3,3}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{4,-4,-4,4}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{4,-4,-4,4}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{4,-4,-4,4}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    {{5,-5,-5,5}, &hhcjets::_M_processA_qQqQ, &hhcjets::_S_select_colorA_QqQq},
    {{5,-5,-5,5}, &hhcjets::_M_processB_qQqQ, &hhcjets::_S_select_colorB_QqQq},
    {{5,-5,-5,5}, &hhcjets::_M_processC_qQqQ, &hhcjets::_S_select_colorC_QqQq},
    
    /*    Process 197-199:   d + d --> d + d
     *    Process 200-202:   u + u --> u + u
     *    Process 203-205:   s + s --> s + s
     *    Process 206-208:   c + c --> c + c
     *    Process 209-211:   b + b --> b + b
     */
    {{-1,-1,1,1}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-1,-1,1,1}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-1,-1,1,1}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-2,-2,2,2}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-2,-2,2,2}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-2,-2,2,2}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-3,-3,3,3}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-3,-3,3,3}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-3,-3,3,3}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-4,-4,4,4}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-4,-4,4,4}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-4,-4,4,4}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    {{-5,-5,5,5}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_qqqq},
    {{-5,-5,5,5}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_qqqq},
    {{-5,-5,5,5}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_qqqq},
    
    /*    Process 212-214:   dbar + dbar --> dbar + dbar
     *    Process 215-217:   ubar + ubar --> ubar + ubar
     *    Process 218-210:   sbar + sbar --> sbar + sbar
     *    Process 221-223:   cbar + cbar --> cbar + cbar
     *    Process 224-226:   bbar + bbar --> bbar + bbar
     */
    {{1,1,-1,-1}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{1,1,-1,-1}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{1,1,-1,-1}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{2,2,-2,-2}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{2,2,-2,-2}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{2,2,-2,-2}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{3,3,-3,-3}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{3,3,-3,-3}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{3,3,-3,-3}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{4,4,-4,-4}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{4,4,-4,-4}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{4,4,-4,-4}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    {{5,5,-5,-5}, &hhcjets::_M_processA_qqqq, &hhcjets::_S_select_colorA_QQQQ},
    {{5,5,-5,-5}, &hhcjets::_M_processB_qqqq, &hhcjets::_S_select_colorB_QQQQ},
    {{5,5,-5,-5}, &hhcjets::_M_processC_qqqq, &hhcjets::_S_select_colorC_QQQQ},
    
    /*  End of list
     */
    {{0,0,0,0}, nullptr, nullptr}
  };
  
  
  /*  Tables for the color supression indices
   */
  const unsigned int hhcjets::_M_color_suppr_idx_4g[6][6] = {
    {0U, 2U, 2U, 2U, 2U, 2U},
    {2U, 0U, 2U, 2U, 2U, 2U},
    {2U, 2U, 0U, 2U, 2U, 2U},
    {2U, 2U, 2U, 0U, 2U, 2U},
    {2U, 2U, 2U, 2U, 0U, 2U},
    {2U, 2U, 2U, 2U, 2U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_2q2g[2][2] = {
    {0U, 2U}, {2U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_4q[2][2] = {
    {0U, 1U}, {1U, 0U}
  };
  
  const unsigned int hhcjets::_M_color_suppr_idx_4qC[2][2] = {
    {0U, 1U}, {1U, 0U}
  };
  
  
  
  hhcjets::hhcjets(const model_ref& mdl, double Ecm, std::initializer_list<double> pT, double x, int hdr1, int hdr2)
  : _M_mdl{mdl}, _M_s{Ecm*Ecm}, _M_scalefac{x*x}
  {
    /*   incoming hadrons   */
    _M_hadronA = hdr1;
    _M_hadronB = hdr2;

    
    /*    the pdf functions   */
    _M_pdf1 = _M_mdl.pdf(qcd::order::lo, hdr1);
    _M_pdf2 = _M_mdl.pdf(qcd::order::lo, hdr2);
    
    
    /*   p2T ranges  */
    for(auto pTx : pT) _M_p2T.push_back(pTx*pTx);
    std::sort(_M_p2T.begin(), _M_p2T.end());
    
    
    /*     normalized quark masses (we have only top mass)    */
    for(int i = 0; i < 7; i++) {
      _M_nu[i] = 0.0;
      _M_nu1[i] = (std::abs(i) < 6 ? 0.0 : _M_mdl.mass2(i)/_M_s);
    }
  }
  
  
  void hhcjets::generate(random_engine_type& random, shower_history<hhc>& hist)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*   resizing the state, setting the model  */
    auto hard = hist.injecting_a_hard_stage();
    
    hard->state.resize(0,2);
    hard->state.set_model(_M_mdl);
    
    /*    incoming hadrons   */
    double Ecm = std::sqrt(_M_s);
    
    hard->state[hadron(-1)].momentum = {0.0,0.0, 0.5*Ecm, 0.5*Ecm};
    hard->state[hadron( 0)].momentum = {0.0,0.0,-0.5*Ecm, 0.5*Ecm};
    hard->state[hadron(-1)].key = _M_hadronA;
    hard->state[hadron( 0)].key = _M_hadronB;
    
    
    /*   generating the (most of the) event   */
    hard->weights.unitary.main = _M_generate_event(random, hard);
  }
  
  
  void hhcjets::_M_generate_x(random_engine_type& random, double& x, double& weight, double xm, double xp)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
    double w1 = (xp-xm)/(xp*xm), w2 = 2.0*std::log(xp/xm);
    
    x = (rng(random) < w1/(w1+w2) ? xp*xm/(xp-(xp-xm)*rng(random)) : xm*std::pow(xp/xm, rng(random)));
    weight *= w1+w2;
    
    /*    symmetrizing in variable x    */
    if(rng(random) < 0.5) x = 1.0-x;
    weight *= 2.0;
  }
  
 
  double hhcjets::_M_generate_event(random_engine_type& random, shower_history<hhc>::iterator hard)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*   picking the p2T range   */
    unsigned ipT = 0u, spT = _M_p2T.size();
    
    do { ipT = static_cast<unsigned>(spT*rng(random)); } while(ipT == spT);
    
    
    /*    generating variable tau    */
    double tau0 = 4.0*_M_p2T[ipT]/_M_s;
    double tau1 = (ipT+1 == spT ? 1.0 : 4.0*_M_p2T[ipT+1]/_M_s);
    double lt = std::log(tau0), tau, weight = static_cast<double>(spT);
    
    tau = std::exp(std::sqrt(rng(random))*lt);
    weight *= 0.5*lt*lt;
    
    _M_p2Tmin = _M_p2T[ipT];
    _M_p2Tmax = (ipT+1 == spT ? 0.25*_M_s*tau : _M_p2T[ipT+1]);
    
    
    /*    generating variable x     */
    double x = 0.5, xmin = 0.5*(1.0-std::sqrt(1.0-tau0/tau));
    double xmax = (tau > tau1 ? 0.5*(1.0-std::sqrt(1.0-tau1/tau)) : 0.5);
    
    if(rng(random) < 0.5) this->_M_generate_x(random, x, weight, xmin, xmax);
    else this->_M_generate_x(random, x, weight, 1.0-xmax, 1.0-xmin);
    
    weight *= 2.0*x*x*(1.0-x)*(1.0-x);
    
    
    /*    generating variable eta and etb    */
    double eta = std::pow(tau, rng(random));
    double etb = tau/eta, shat = _M_s*tau;
    
    
    /*      the scale of the hard event    */
    hard->scale = _M_scalefac*shat*x*(1.0-x);
    
//    switch(_M_mdl.scheme()) {
//      case ordering::scheme::angular:
//      case ordering::scheme::kT: hard->scale = _M_scalefac*shat*x*(1.0-x); break;
//      case ordering::scheme::lambda: hard->scale = _M_scalefac*shat*std::min(x, 1.0-x); break;
//    }
    
    
    /*    pre-calculating the pdf factors    */
    for(int k = -5; k <= 5; ++k) {
      _M_fa[k+5] = _M_pdf1 -> parton(k, eta, hard->scale);
      _M_fb[k+5] = _M_pdf2 -> parton(k, etb, hard->scale);
    }
  
    
    /*    QCD coupling    */
    double as = 2.0*M_PI*_M_mdl.alpha_qcd(hard->scale);
    
    weight *= 4.0*M_PI*as*as/shat;
    
    
    /* Here we select the flavor configuration.
     * For this we use rather complicated weights those
     * consist of pdfs and some x dependent function.
     */
    weight *= this->_M_select_subprocess(tau, x, rng(random));
    
    hard->state[-1].flavor = _M_active->fl[0];
    hard->state[ 0].flavor = _M_active->fl[1];
    hard->state[ 1].flavor = _M_active->fl[2];
    hard->state[ 2].flavor = _M_active->fl[3];

    
    /*    For the outgoing momenta we use Sudakov paremetrization.   */
    int f1 = hard->state[1].flavor, f2 = hard->state[2].flavor;
    
    double nh1 = (std::abs(f1) < 6 ? 0.0 : _M_mdl.mass2(f1)/shat);
    double nh2 = (std::abs(f2) < 6 ? 0.0 : _M_mdl.mass2(f2)/shat);
    
    /*     +/- and transverse momenta      */
    double phi = 2.0*M_PI*rng(random), mT = std::sqrt((x*(1.0-x) - x*nh2 - (1.0-x)*nh1)*shat);
    
    lorentzvector<double> na = hard->state[hadron(-1)].momentum*eta;
    lorentzvector<double> nb = hard->state[hadron( 0)].momentum*etb;
    lorentzvector<double> kT(mT*std::cos(phi), mT*std::sin(phi), 0.0, 0.0);
    
    
    /*    reconstructing the parton momenta    */
    hard->state[-1].momentum = na;
    hard->state[ 0].momentum = nb;
    hard->state[ 1].momentum = (1.0-x+nh1-nh2)*na +      x *nb + kT;
    hard->state[ 2].momentum =     (x+nh2-nh1)*na + (1.0-x)*nb - kT;
    
    
    /*    selecting the color states    */
    weight *= (*(_M_active->select_color))(hard->state, rng(random));
    
    
    /*   vetos    */
    hard->vetos = shower_vetos{};
    hard->vetos.max_k2T = shat*x*(1.0-x);
    
//    switch(_M_mdl.scheme()) {
//      case ordering::scheme::angular:
//      case ordering::scheme::kT: hard->vetos.max_k2T = shat*x*(1.0-x); break;
//      case ordering::scheme::lambda: hard->vetos.max_q2 = shat*std::min(x, 1.0-x); break;
//    }
   
    
    return weight;
  }
  
  
  double hhcjets::_M_select_subprocess(double tau, double x, double rho)
  {
    /*     calculate the weight of the subprocesses    */
    double weight[250], norm = 0.0;
    double *iwgt = weight;
    
    for(auto iproc = _M_proc_tbl; iproc->weight_function; ++iproc, ++iwgt)
      norm += (*iwgt = (this ->* (iproc->weight_function))(tau,x, iproc->fl));
    
    
    /*     selecting subprocess     */
    unsigned ip = 0U;
    double cproc = weight[0];
    
    while(cproc < rho*norm)
      cproc += weight[++ip];
    
    /*    active subprocess     */
    _M_active = _M_proc_tbl + ip;
    
    
    return norm;
  }
  
  
  
  /*  Process 0: g + g --> g + g
   */
  double hhcjets::_M_process_gggg(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != 0 || fl[1] != 0 || fl[2] != 0 || fl[3] != 0)
      throw "hhcjets::_M_process_gggg(....) : wrong flavor";
#endif
    
    constexpr double cfac = qcd::Na2/(qcd::Nc2*qcd::Nc2);
    
    //----- transverse momentum check -----
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(!(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight ----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox2 = 1.0-x;
    ox2 *= ox2;
    
    return fa[0]*fb[0]*cfac*(1.0+x2*x2+ox2*ox2)/(x2*ox2);
  }
  
  
  double hhcjets::_S_select_color_gggg(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x/16.0, p12 = x*ox/16.0, p13 = x/16.0, p22 = ox*ox/16.0, p23 = ox/16.0, p33 = 1.0/16.0;
    
    const double color_prob[6][6] = {
      {p11, p12, p13, p11, p12, p13},
      {p12, p22, p23, p12, p22, p23},
      {p13, p23, p33, p13, p23, p33},
      {p11, p12, p13, p11, p12, p13},
      {p12, p22, p23, p12, p22, p23},
      {p13, p23, p33, p13, p23, p33}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w13 = -1.0, w22 = 1.0, w23 = -1.0, w33 = 1.0;
    
    static const double color_weight[6][6] = {
      {w11, w12, w13, w11, w12, w13},
      {w12, w22, w23, w12, w22, w23},
      {w13, w23, w33, w13, w23, w33},
      {w11, w12, w13, w11, w12, w13},
      {w12, w22, w23, w12, w22, w23},
      {w13, w23, w33, w13, w23, w33}
    };
    
    //----- all the possible color state -----
    static const int perm[6][4] = {
      {-1,0,1,2}, {-1,1,2,0}, {-1,2,0,1}, {2,1,0,-1}, {0,2,1,-1}, {1,0,2,-1}
    };
    
    //----- selecting color state ----
    const double *cp = (const double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/6U, j = ic%6U;
    
    //----- inserting the color strings -----
    state.insert_closed_string(perm[i], perm[i]+4);
    state.insert_closed_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4g[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 1-6: g + g --> q + qbar, where q = d,u,s,c,b,t
   */
  double hhcjets::_M_process_ggqQ(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != 0 || fl[1] != 0 || fl[2] != -fl[3] || fl[2] == 0)
      throw "hhcjets::_M_process_ggqQ(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(!(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox = 1.0-x;
    double ox2 = ox*ox;
    
    //----- leading contribution -----
    double retval = x2 + ox2;
    
    //----- extra term for heavy quarks -----
    double nu = _M_nu[std::abs(fl[2])];
    
    if(nu > 1e-14) {
      double ni = nu/tau;
      retval += 4.0*ni*(1.0 - ni/(x*ox));
    }
    
    return fa[0]*fb[0]*retval/(8.0*qcd::Nc*x*ox);
  }
  
  
  double hhcjets::_S_select_color_ggqQ(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = ox*ox, p12 = x*ox, p22 = x*x;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,-1,0,2}, {1,0,-1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process  7-11: g + q --> g + q, where q = d,u,s,c,b
   *    Process 12-16: g + qbar --> g + qbar where q = d,u,s,c,b
   */
  double hhcjets::_M_process_gqgq(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != 0 || fl[2] != 0 || fl[1] != -fl[3] || fl[1] == 0)
      throw "hhcjets::_M_process_gqgq(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(!(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double x2 = x*x, ox = 1.0-x;
    double ox2 = ox*ox;
    int ib = -fl[1];
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na/(8.0*qcd::Nc2);
    double retval = (1.0+ox2)/(x2*ox);
    
    //----- extra term for heavy quarks -----
    double nu = _M_nu[std::abs(ib)];
    
    if(nu > 1e-14) {
      double ni = nu/(tau*ox);
      retval -= 4.0*ni*(1.0/x - ni);
    }
    
    retval *= fa[0]*fb[ib]*cfac;
    retval *= (2.0-x)*(2.0-x);         //  this is part of the color weights in the notes
    
    return retval;
  }
  
  
  double hhcjets::_S_select_color_gqgq(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {2,-1,1,0}, {2,1,-1,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_S_select_color_gQgQ(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,1,-1,2}, {0,-1,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 17-21: q + g --> g + q, where q = d,u,s,c,b
   *    Process 22-26: qbar + g --> g + qbar, where q = d,u,s,c,b
   */
  double hhcjets::_M_process_qggq(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[1] != 0 || fl[2] != 0 || fl[0] != -fl[3] || fl[0] == 0)
      throw "hhcjets::_M_process_qggq(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(!(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    int ia = -fl[0];
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double nu = _M_nu[std::abs(ia)];
    double y = tau/(tau+nu) - x;
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na/(8.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(y*y*oy);
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = nu/(tau*oy);
      retval -= 4.0*ni*(1.0/y - ni);
    }
    
    retval *= fa[ia]*fb[0]*cfac;
    retval *= (2.0-y)*(2.0-y);      //  this is part of the color weights in the notes
    
    return retval;
  }
  
  
  double hhcjets::_S_select_color_qggq(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[0].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {2,0,1,-1}, {2,1,0,-1}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_S_select_color_QggQ(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[0].momentum*state[1].momentum)/(state[0].momentum*state[-1].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x, tx2 = (2.0-x)*(2.0-x);
    double p11 = 1.0/tx2, p12 = ox/tx2, p22 = ox*ox/tx2;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,0,2}, {-1,0,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 27-31: q + qbar --> g + g, where q = d,u,s,c,b
   *    Process 32-36: qbar + q --> g + g, where q = d,u,s,c,b
   */
  double hhcjets::_M_process_qQgg(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[2] != 0 || fl[3] != 0 || fl[0] != -fl[1] || fl[0] == 0)
      throw "hhcjets::_M_process_qQgg(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (nu + x*(tau-nu))/(tau+nu);
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = qcd::Na2/(16.0*qcd::Nc2*qcd::Nc);
    double retval = y*y+oy*oy;
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = nu*tau/(tau+nu)/(tau+nu);
      retval += 4.0*ni*(1.0-ni/(y*oy));
    }
    
    retval *= cfac/(y*oy);
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  
  double hhcjets::_S_select_color_qQgg(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[1].momentum*state[2].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x, p12 = x*ox, p22 = ox*ox;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,1,2,-1}, {0,2,1,-1}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_S_select_color_Qqgg(event<hhc>& state, double rho)
  {
    //----- the kinematical variable -----
    double x = (state[-1].momentum*state[1].momentum)/(state[1].momentum*state[2].momentum);
    
    //----- color probabilities -----
    double ox = 1.0-x;
    double p11 = x*x, p12 = x*ox, p22 = ox*ox;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = 1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,1,0}, {-1,1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+4);
    state.insert_open_stringP(perm[j], perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_2q2g[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 37-41: d + dbar --> r + rbar, where r =   u,s,c,b,t
   *    Process 42-46: u + ubar --> r + rbar, where r = d,  s,c,b,t
   *    Process 47-51: s + sbar --> r + rbar, where r = d,u,  c,b,t
   *    Process 52-56: c + cbar --> r + rbar, where r = d,u,s,  b,t
   *    Process 57-61: b + bbar --> r + rbar, where r = d,u,s,c,  t
   
   *    Process 62-66: dbar + d --> rbar + r, where r =   u,s,c,b,t
   *    Process 67-71: ubar + u --> rbar + r, where r = d,  s,c,b,t
   *    Process 72-76: sbar + s --> rbar + r, where r = d,u,  c,b,t
   *    Process 77-81: cbar + c --> rbar + r, where r = d,u,s,  b,t
   *    Process 81-86: bbar + b --> rbar + r, where r = d,u,s,c,  t
   */
  double hhcjets::_M_process_qQrR(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[1] || fl[2] != -fl[3] || fl[0] == 0 || fl[2] == 0)
      throw "hhcjets::_M_process_qQrR(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0], i1 = fl[2];
    double nu = _M_nu[std::abs(ia)], nu1 = _M_nu1[std::abs(i1)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (nu + x*(tau-nu))/(tau+nu);
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(8.0*qcd::Nc2);
    double retval = y*y+oy*oy;
    
    //----- extra term for heavy quarks -----
    if(nu + nu1 > 1e-14)
      retval += 2.0*tau*(nu+nu1)/(tau+nu)/(tau+nu);
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_color_qQrR(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_color_QqRr(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {2,1,-1,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  
  /*    Process 87-90:   d + r --> d + r, where r =   u,s,c,b
   *    Process 91-94:   u + r --> u + r, where r = d,  s,c,b
   *    Process 95-98:   s + r --> s + r, where r = d,u,  c,b
   *    Process 99-102:  c + r --> c + r, where r = d,u,s,  b
   *    Process 103-106: b + r --> b + r, where r = d,u,s,c
   
   *    Process 107-110: dbar + rbar --> dbar + rbar, where r =   u,s,c,b
   *    Process 111-114: ubar + rbar --> ubar + rbar, where r = d,  s,c,b
   *    Process 115-118: sbar + rbar --> sbar + rbar, where r = d,u,  c,b
   *    Process 119-122: cbar + rbar --> cbar + rbar, where r = d,u,s,  b
   *    Process 123-126: bbar + rbar --> bbar + rbar, where r = d,u,s,c
   
   *    Process 127-130: d + rbar --> d + rbar, where r =   u,s,c,b
   *    Process 131-134: u + rbar --> u + rbar, where r = d,  s,c,b
   *    Process 135-138: s + rbar --> s + rbar, where r = d,u,  c,b
   *    Process 139-142: c + rbar --> c + rbar, where r = d,u,s,  b
   *    Process 143-146: b + rbar --> b + rbar, where r = d,u,s,c
   
   *    Process 147-150: dbar + r --> dbar + r, where r =   u,s,c,b
   *    Process 151-154: ubar + r --> ubar + r, where r = d,  s,c,b
   *    Process 155-158: sbar + r --> sbar + r, where r = d,u,  c,b
   *    Process 159-162: cbar + r --> cbar + r, where r = d,u,s,  b
   *    Process 163-166: bbar + r --> bbar + r, where r = d,u,s,c
   */
  double hhcjets::_M_process_qrqr(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || std::abs(fl[0]) == std::abs(fl[1]) || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_process_qrqr(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0], ib = -fl[1];
    double nua = _M_nu[std::abs(ia)], nub = _M_nu[std::abs(ib)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < std::sqrt(nua*nub) || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau*tau-nua*nub)/(tau*tau+nua*nub)*(x-nua/(tau+nua));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(8.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(y*y);
    
    //----- extra term for heavy quarks -----
    if(nua + nub > 1e-14)
      retval -= 2.0*tau/y*(nua+nub)/(tau*tau+nua*nub);
    
    retval *= cfac;
    retval *= fa[ia]*fb[ib];
    
    return retval;
  }
  
  double hhcjets::_S_select_color_qrqr(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_color_QRQR(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_color_qRqR(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,2,0,-1}, {1,-1,0,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_color_QrQr(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,0,2,1}, {-1,1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 167-169:   d + dbar --> d + dbar
   *    Process 170-172:   u + ubar --> u + ubar
   *    Process 173-175:   s + sbar --> s + sbar
   *    Process 176-168:   c + cbar --> c + cbar
   *    Process 179-181:   b + bbar --> b + bbar
   
   *    Process 182-184:   dbar + d --> dbar + d
   *    Process 185-187:   ubar + u --> ubar + u
   *    Process 188-190:   sbar + s --> sbar + s
   *    Process 191-193:   cbar + c --> cbar + c
   *    Process 194-196:   bbar + b --> bbar + b
   */
  double hhcjets::_M_processA_qQqQ(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != -fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processA_qQqQ(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau-nu)/(tau+nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = y*y+oy*oy;
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau+nu)/(tau+nu);
      retval += 8.0*ni*(y+ni);
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorA_qQqQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorA_QqQq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_M_processB_qQqQ(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != -fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processB_qQqQ(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau-nu)/(tau+nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(y*y);
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau+nu)/(tau+nu);
      retval -= 8.0*ni*(1.0-ni)/(y*y);
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorB_qQqQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorB_QqQq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_M_processC_qQqQ(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != -fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processC_qQqQ(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau-nu)/(tau+nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = 2.0*oy*oy/y;
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau+nu)/(tau+nu);
      retval -= 8.0*ni*ni/y;
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[-ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorC_qQqQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {0,2,1,-1}, {0,-1,1,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorC_QqQq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,1,2,0}, {-1,0,2,1}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    
    return color_weight[i][j];
  }
  
  
  /*    Process 197-199:   d + d --> d + d
   *    Process 200-202:   u + u --> u + u
   *    Process 203-205:   s + s --> s + s
   *    Process 206-208:   c + c --> c + c
   *    Process 209-211:   b + b --> b + b
   
   *    Process 212-214:   dbar + dbar --> dbar + dbar
   *    Process 215-217:   ubar + ubar --> ubar + ubar
   *    Process 218-210:   sbar + sbar --> sbar + sbar
   *    Process 221-223:   cbar + cbar --> cbar + cbar
   *    Process 224-226:   bbar + bbar --> bbar + bbar
   */
  double hhcjets::_M_processA_qqqq(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processA_qqqq(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau+nu)/(tau-nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+oy*oy)/(y*y);
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau-nu)/(tau-nu);
      retval += 8.0*ni*(1.0-y+ni)/(y*y);
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorA_qqqq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorA_QQQQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p22}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_M_processB_qqqq(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processB_qqqq(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau+nu)/(tau-nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = (1.0+y*y)/(oy*oy);
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau-nu)/(tau-nu);
      retval += 8.0*ni*(y+ni)/(oy*oy);
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorB_qqqq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorB_QQQQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double p22 = 1.0/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = qcd::Nc2*p22, p12 = qcd::Nc*p22;
    
    const double color_prob[2][2] = {
      {p22, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {w11, w12},
      {w12, w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4q[i][j];
    
    return color_weight[i][j];
  }
  
  
  double hhcjets::_M_processC_qqqq(double tau, double x, const int *fl) const
  {
#ifdef __hhcjets_CHECK_FLAVORS__
    //----- flavor check -----
    if(fl[0] != -fl[2] || fl[1] != -fl[3] || fl[0] != fl[1] || fl[0] == 0 || fl[1] == 0)
      throw "hhcjets::_M_processC_qqqq(....) : wrong flavor";
#endif
    
    //----- transverse momentum check -----
    int ia = -fl[0];
    double nu = _M_nu[std::abs(ia)];
    
    double p2T = this->_M_transverse_momentum(tau, x, fl);
    if(tau < nu || !(_M_p2Tmin < p2T && p2T < _M_p2Tmax)) return 0.0;
    
    //----- calculate the weight -----
    const double *fa = _M_fa+5, *fb = _M_fb+5;
    double y = (tau+nu)/(tau-nu)*(x-nu/(tau+nu));
    double oy = 1.0-y;
    
    //----- leading contribution -----
    constexpr double cfac = (qcd::Nc+1.0)*(qcd::Nc+1.0)/(16.0*qcd::Nc2);
    double retval = 2.0/(y*oy);
    
    //----- extra term for heavy quarks -----
    if(nu > 1e-14) {
      double ni = tau*nu/(tau-nu)/(tau-nu);
      retval -= 8.0*ni*ni/(y*oy);
    }
    
    retval *= cfac;
    retval *= fa[ia]*fb[ia];
    
    return retval;
  }
  
  double hhcjets::_S_select_colorC_qqqq(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {-w11, -w12},
      {-w12, -w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {1,0,2,-1}, {1,-1,2,0}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    
    return color_weight[i][j];
  }
  
  double hhcjets::_S_select_colorC_QQQQ(event<hhc>& state, double rho)
  {
    //----- color probabilities -----
    double pp = 0.5/((qcd::Nc+1.0)*(qcd::Nc+1.0));
    double p11 = 2.0*qcd::Nc*pp, p12 = (qcd::Nc2+1.0)*pp;
    
    const double color_prob[2][2] = {
      {p11, p12},
      {p12, p11}
    };
    
    //----- color weights -----
    constexpr double w11 = 1.0, w12 = -1.0, w22 = 1.0;
    
    static const double color_weight[2][2] = {
      {-w11, -w12},
      {-w12, -w22}
    };
    
    //----- all the possible color state -----
    static const int perm[2][4] = {
      {-1,2,0,1}, {-1,1,0,2}
    };
    
    //----- selecting color state ----
    double *cp = (double *) color_prob;
    double cprob = cp[0];
    unsigned int ic = 0U;
    
    while(cprob < rho)
      cprob += cp[++ic];
    
    unsigned int i = ic/2U, j = ic%2U;
    
    //----- inserting the color strings -----
    state.insert_open_string(perm[i], perm[i]+2);
    state.insert_open_string(perm[i]+2, perm[i]+4);
    
    state.insert_open_stringP(perm[j], perm[j]+2);
    state.insert_open_stringP(perm[j]+2, perm[j]+4);
    
    //----- setting the color suppression index of the selected configuration -----
    state.color_suppression_index = _M_color_suppr_idx_4qC[i][j];
    
    return color_weight[i][j];
  }
  
  
}


